#define STRICT
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <condefs.h>
#pragma hdrstop
//---------------------------------------------------------------------------
HINSTANCE hInstance;
HWND hwnd;
HWND hed;

bool strict=false;

char errbuf[1000];
char *err()
{ FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM,0,GetLastError(),
      MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
      errbuf,0,NULL);
  return errbuf;
}

char *find_fn(char *pathfn)
{ char *fn1=strrchr(pathfn,'\\');
  char *fn2=strrchr(pathfn,':');
  if (fn1>fn2) return fn1+1;
  if (fn2>0) return fn2+1;
  return pathfn;
}

void print(char *msg)
{ char c[1000]; strcpy(c,msg); strcat(c,"\r\n");
  SendMessage(hed,EM_REPLACESEL,FALSE,(LPARAM)c);
  SendMessage(hed,EM_SETSEL,-1,-1);
}


void FixDns(char *fn)
{ HANDLE hfdns, hmdns; char *pdns;

  char *fnbase = find_fn(fn); char msg[2000];
  wsprintf(msg,"** %s",fnbase); print(msg);
  // Now, open the DNS file
  hfdns = CreateFile(fn,GENERIC_READ|GENERIC_WRITE,0,NULL,OPEN_EXISTING,0,NULL);
  if (hfdns==INVALID_HANDLE_VALUE) {wsprintf(msg,"couldn't open file -- %s",err()); print(msg); return;}
  DWORD hisize=0; DWORD size = GetFileSize(hfdns,&hisize);
  if (hisize!=0) {wsprintf(msg,"superlarge files not supported"); print(msg); CloseHandle(hfdns); return;}
  hmdns = CreateFileMapping(hfdns,NULL,PAGE_READWRITE,0,0,NULL);
  if (hmdns==NULL) {wsprintf(msg,"couldn't map file -- %s",err()); print(msg); CloseHandle(hfdns); return;}
  pdns = (char*)MapViewOfFile(hmdns,FILE_MAP_ALL_ACCESS,0,0,0);
  if (pdns==NULL) {wsprintf(msg,"couldn't lock file -- %s",err()); print(msg); CloseHandle(hmdns); CloseHandle(hfdns); return;}
  unsigned int src=0, dst=0;

  // The format is:
  // 16673366 1.0.0.1 *
  // 16641907 62.11.0.1 pi1-1.dialup.tiscalinet.it
  // number <blanks> num.num.num.num <blanks> name <blanks> <cr>
  // The error we'll catch is anything other than this. usually
  // because "name" has a space in it.
  unsigned int op=0; int nlines=0; unsigned int starttime=GetTickCount();
  char *cause;
  while (src<size && pdns[src]!=0)
  { nlines++;
    // we have a line starting at src. If it's okay, we'll copy
    // it into dstline. If it's bad, we'll pass.
    int i=0; bool okay=true; int dnsstart=0;
    char *line = pdns+src;
    if (line==NULL) MessageBeep(0);
    // 0. an empty line
    bool blank=false;
    if (pdns[src+i]=='\r' && pdns[src+i+1]=='\n') {i+=2; blank=true;}
    else if (pdns[src+i]=='\n') {i++; blank=true;}
    if (!blank)
    { // 1. a number
      if (okay)
      { int start=i;
        while (pdns[src+i]>='0' && pdns[src+i]<='9') i++;
        if (i==start) {okay=false; cause="missing number at start";}
      }
      // 2. blanks
      if (okay)
      { int start=i;
        while (pdns[src+i]==' ' || pdns[src+i]=='\t') i++;
        if (i==start) {okay=false; cause="missing space after initial number";}
      }
      // 3. num dot num dot num dot num
      for (int ip=0; ip<4; ip++)
      { if (okay)
        { int start=i;
          while (pdns[src+i]>='0' && pdns[src+i]<='9') i++;
          if (i==start) {okay=false; cause="misformed ip number, missing digit";}
        }
        if (okay && ip!=3)
        { if (pdns[src+i]=='.') i++; else {okay=false; cause="misformed ip number, missing dot";}
        }
      }
      // 4. blanks
      if (okay)
      { int start=i;
        while (pdns[src+i]==' ' || pdns[src+i]=='\t') i++;
        if (i==start) {okay=false; cause="missing space after ip";}
      }
      // 5. dns name
      if (okay)
      { dnsstart=i;
        if (pdns[src+i]=='*') i++;
        else
        { int start=i;
          if (strict)
          { while ((pdns[src+i]>='a' && pdns[src+i]<='z') || (pdns[src+i]>='A' && pdns[src+i]<='Z') || (pdns[src+i]>='0' && pdns[src+i]<='9')
                 || pdns[src+i]=='-' || pdns[src+i]=='.') i++;
          }
          else
          { while ((pdns[src+i]>='a' && pdns[src+i]<='z') || (pdns[src+i]>='A' && pdns[src+i]<='Z') || (pdns[src+i]>='0' && pdns[src+i]<='9')
                 || pdns[src+i]=='-' || pdns[src+i]=='.' || pdns[src+i]=='#' || pdns[src+i]=='@'
                 || pdns[src+i]=='_' || pdns[src+i]=='~' || pdns[src+i]=='/' || pdns[src+i]=='*') i++;
          }
          if (i==start) {okay=false; cause="missing dns name";}
        }
      }
      // 6. possibly terminal blanks
      if (okay)
      { while (pdns[src+i]==' ' || pdns[src+i]=='\t') i++;
      }
      // 7. newline
      if (okay)
      { if (pdns[src+i]=='\r' && pdns[src+i+1]=='\n') i+=2;
        else if (pdns[src+i]=='\n') i+=1;
        else {okay=false; cause="illegal space in dns name";}
      }
    }

    if (okay)
    { if (src!=dst) CopyMemory(pdns+dst,pdns+src,i);
      src+=i;
      dst+=i;
    }
    else
    { while (pdns[src+i]!=0 && pdns[src+i]!='\r' && pdns[src+i]!='\n') i++;
      bool crlf = (pdns[src+i]=='\r');
      if (pdns[src+i]!=0) i++;
      char c[100];
      if (i>=100) {CopyMemory(c,pdns+src,100); c[100]=0;}
      else {CopyMemory(c,pdns+src,i); c[i]=0;}
      while (pdns[src+i]=='\n') i++;
      wsprintf(msg,"%s -- %s",c,cause); print(msg);
      // but maybe we can salvage something? If the dns itself was bad...
      if (dnsstart!=0)
      { int j=0;
        while (j<dnsstart) {pdns[dst+j]=pdns[src+j]; j++;}
        pdns[dst+j]='*'; j++;
        if (crlf) {pdns[dst+j]='\r'; pdns[dst+j+1]='\n'; j+=2;}
        else {pdns[dst+j]='\n'; j++;}
        dst += j;
      }
      src += i;
    }

    if (10*src/size>op)
    { op=10*src/size;
      wsprintf(msg,"%s - %i%s",fnbase,op*10,"%");
      SetWindowText(hwnd,msg);
    }
  }

  unsigned int endtime=GetTickCount();
  wsprintf(msg,"Finished. %i lines, %ik, %ims",nlines,size/1024, endtime-starttime);
  print(msg);
  SetWindowText(hwnd,"Fix DNS");
  UnmapViewOfFile(pdns);
  CloseHandle(hmdns);
  if (dst!=size)
  { SetFilePointer(hfdns,dst,0,FILE_BEGIN);
    SetEndOfFile(hfdns);
  }
  CloseHandle(hfdns);
}

LRESULT CALLBACK PlainWndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{ switch (msg)
  { case WM_CREATE:
    { hed = CreateWindow("EDIT","",WS_CHILD|WS_VISIBLE|WS_HSCROLL|WS_VSCROLL|ES_LEFT|ES_MULTILINE,
                    0,0,100,100,hwnd,NULL,hInstance,0);
      print("Drop your dns file here. I will remove any bad lines.");
      print("(you might wish to backup the dns file first...)");
      if (strict) print("Running in strict mode (no punctuation except hyphens)");
      else print("Running in lax mode (punctuation allowed)");
      DragAcceptFiles(hwnd,TRUE);
    } break;
    case WM_SIZE:
    { RECT rc; GetClientRect(hwnd,&rc);
      MoveWindow(hed,0,0,rc.right,rc.bottom,TRUE);
    } break;
    case WM_PAINT:
    { PAINTSTRUCT ps; BeginPaint(hwnd,&ps);
      EndPaint(hwnd,&ps);
    } break;
    case WM_DROPFILES:
    { HDROP hdrop = (HDROP)wParam;
      int num = DragQueryFile(hdrop,0xFFFFFFFF,NULL,0);
      for (int i=0; i<num; i++)
      { int len = DragQueryFile(hdrop,i,0,0);
        char *c=new char[len+1];
        DragQueryFile(hdrop,i,c,len+1);
        FixDns(c);
      }
      DragFinish(hdrop);
    } break;
    case WM_CLOSE:
    { DestroyWindow(hwnd);
    } break;
    case WM_DESTROY:
    { DragAcceptFiles(hwnd,FALSE);
      PostQuitMessage(0);
    } break;
  }
  return DefWindowProc(hwnd, msg, wParam, lParam);
}



int WINAPI WinMain(HINSTANCE h,HINSTANCE,LPSTR,int)
{ hInstance=h;
  //
  WNDCLASSEX wcex; ZeroMemory(&wcex,sizeof(wcex)); wcex.cbSize = sizeof(WNDCLASSEX);
  BOOL res=GetClassInfoEx(hInstance,"PlainClass",&wcex);
  if (!res)
  { wcex.style = CS_HREDRAW | CS_VREDRAW;
    wcex.lpfnWndProc = (WNDPROC)PlainWndProc;
    wcex.cbClsExtra = 0;
    wcex.cbWndExtra = 0;
    wcex.hInstance = hInstance;
    wcex.hIcon = NULL;
    wcex.hCursor = LoadCursor(NULL, IDC_ARROW);
    wcex.hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH);
    wcex.lpszMenuName	= NULL;
    wcex.lpszClassName = "PlainClass";
    wcex.hIconSm = NULL;
    ATOM res=RegisterClassEx(&wcex);
    if (res==0) {MessageBox(NULL,"Failed to register class","Error",MB_OK); return 0;}
  }
  //
  hwnd = CreateWindowEx(0,"PlainClass", "Fix DNS", WS_OVERLAPPEDWINDOW,
      CW_USEDEFAULT, CW_USEDEFAULT, 400, 400, NULL, NULL, hInstance, NULL);
  if (hwnd==NULL) {MessageBox(NULL,"Failed to create window","Error",MB_OK); return 0;}
  ShowWindow(hwnd,SW_SHOW);
  //
  MSG msg;
  while (GetMessage(&msg, NULL, 0, 0))
  { TranslateMessage(&msg);
	DispatchMessage(&msg);
  }
  return msg.wParam;
}

 
